---------------------------------------------------------------------------
License 
Copyright 2004 Grigory (Greg) Nudelman. All Rights Reserved. 
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: 

Redistribution of source code must retain this copyright notice, this list of conditions and the following disclaimer.

Redistribution in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. 

The name of Grigory (Greg) Nudelman may not be used to endorse or promote products derived from this software without specific prior written permission. 

This software is provided "AS IS," without a warranty of any kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. THE INVENTOR, Grigory (Greg) Nudelman AND THE LICENSORS OF THIS SOFTWARE SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL THE INVENTOR OR HIS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF THE INVENTOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. 

You acknowledge that this software is not designed, licensed or intended for use in the design, construction, operation or maintenance of any nuclear facility.  
---------------------------------------------------------------------------

package cache;

import java.io.*;
import java.sql.*;
import java.util.*;

/**
 * This class models a CD (or any other object for that matter) 
 * that you have to create on-demand from one or more JDBC calls.
 * CD comes with the attached time stamp that changes when the 
 * object is refreshed in the database.
 * CD is used in the cache demo.
 * Note: This class (and the entire demo as well) 
 * assumes that the id used in the system is a String.
 * If it's not, please adjust the code accordingly.
 *
 * @author  Greg Nudelman
 * @version 1.0
 * @see     Cache
 */
public class CD implements Cacheable {

    private String id;
    private String title;
    private String artist;
    private String imageUrl;
    private int remainingInInventory;
    private double price;
    private long timestamp;

    /**
     * Performs simplistic version of the initial search in the DB
     * to retrieve all CD ids and timestamps that 
     * match the user's search criteria.
     * For example search for "Britney" by title should return all 
     * the ids/timestamps for the Britney Spears CDs.
     * Because all of the CD search and retrieval logic is likely
     * to be performed in a single transaction, we'll be passing 
     * in the JDBC Connection object to run the SQL.
     * 
     * @return List of TimestampedIds that match the user's search.
     * This vector can then be used to retrieve pre-loaded CDs from 
     * the Cache using the standard API.
     *
     * @see TimestampedId, Cache
     * @exception ResourceLoadException in case of critical error
     * @author Greg Nudelman <GN>
     * @version 1.0
     */
    public static List getCdIdsThatMatchUserSearch(String artist,
                                                    Connection conn) throws ResourceLoadException {        
        if(null == artist)
            throw new IllegalArgumentException("Artist name can't be null.");

        Vector returnVec = new Vector(); //return a vector of timestamped Id objects

        System.out.println("getCdIdsThatMatchUserSearch(" + artist + ")");

        String sql = new StringBuffer()
            .append("SELECT id, timestamp ") 
			//.append("FROM cd WHERE upper(artist) LIKE '%") //oracle SQL
            .append("FROM cd WHERE ucase(artist) LIKE '%") //access SQL 
			//Weird: Access requires '*B*', but that does not work? Also: ucase() not needed? 
            .append(artist.toUpperCase())
            .append("%'")
            .toString();

        System.out.println("getCdIdsThatMatchUserSearch() sql = " + sql);

        Statement stmt = null;
        ResultSet rs = null;

        try {
            stmt = conn.createStatement();                               
            rs = stmt.executeQuery(sql);         

            TimestampedId id = null;
            while (rs.next()) {
                id = new TimestampedId(rs.getString(1), rs.getLong(2));
                System.out.println("getCdIdsThatMatchUserSearch() id = " + id);
                returnVec.addElement(id);
            } 

        } catch(SQLException sqlex) {
            throw new ResourceLoadException(sqlex.getMessage());

        } catch(Exception ex) {
            throw new ResourceLoadException(ex.getMessage());

        } finally {
            try {
                if(rs != null) {
                    rs.close();
                    rs = null;
                }
            } catch (Exception ignored1) {
                //ignored
                System.out.println("WARNING: rs.close() Exception: " + ignored1);
                ignored1.printStackTrace();
            }

            try {
                if(stmt != null) {
                    stmt.close();
                    stmt = null;
                }
            } catch (Exception ignored2) {
                //ignored
                System.out.println("WARNING: stmt.close() Exception: " + ignored2);
                ignored2.printStackTrace();
            }
            //do not close connection: re-use for the next SQL in the transaction
        }

        return returnVec;
    }



    ////////////////////// CACHE MANAGER AND CACHEABLE METHODS ///////////////////////

    /**
     * method to demonstrate object loading using only the <code>Timestamped</code> methods
     * and Reflection API
     * 
     * @see CacheManager.loadFromCacheAndDB(String String)
     */
    public static Collection loadFromIdsUsingReflection(Collection idsToRefreshFromDb, 
                                                        Connection conn) throws ResourceLoadException {
        return (new CD()).loadFromIds(idsToRefreshFromDb, conn);
    }


    /**
     * Method needed to implement Cacheable interface.
     * For the passed Collection of String ids creates new CD objects and 
     * loads them with all the attributes from the DB.
     * The resulting objects can be cached to be re-used later.
     *
     * NOTE: comes back as unordered as a Collection!!
     * Make sure you sort the final List of cached + db objects BEFORE you return it!
     *
     * @return Collection of fully loaded CD objects from the DB.
     * @author Greg Nudelman
     * @exception ResourceLoadException in case of critical error
     * @version 1.0
     */
    public Collection loadFromIds(Collection idsToRefreshFromDb, 
                                  Connection conn) throws ResourceLoadException {
        
        if(idsToRefreshFromDb.isEmpty()) {
            System.out.println("loadFromDB() nothing to load. Returning empty.");
            return new Vector(1);
        }

        Vector objectsFromDb = new Vector(idsToRefreshFromDb.size()); //result

        String commaDelimitedIds = Util.makeCommaDelimitedStringsList(idsToRefreshFromDb);
        System.out.println("loadFromDB(" + commaDelimitedIds + ")");


        String sql = new StringBuffer()
            .append("SELECT id, title, artist, imageUrl, remainingInInventory, price, timestamp ")
            .append("FROM cd WHERE id IN (")
            .append(commaDelimitedIds)
            .append(")")
            .toString(); 
        System.out.println("loadFromDB() sql = " + sql);

        Statement stmt = null;
        ResultSet rs = null;

        try {
            stmt = conn.createStatement();                               
            rs = stmt.executeQuery(sql);         

            CD cd = null;
            while (rs.next()) {
                cd = new CD();
                cd.id                   = rs.getString(1);
                cd.title                = rs.getString(2);
                cd.artist               = rs.getString(3);
                cd.imageUrl             = rs.getString(4);
                cd.remainingInInventory = rs.getInt(5);
                cd.price                = rs.getDouble(6);
                cd.timestamp            = rs.getLong(7);
                
                System.out.println("loadFromDB() cd = " + cd);
                objectsFromDb.addElement(cd);
            } 

        } catch(SQLException sqlex) {
            throw new ResourceLoadException(sqlex.getMessage());

        } catch(Exception ex) {
            throw new ResourceLoadException(ex.getMessage());

        } finally {
            try {
                if(rs != null)
                    rs.close();
            } catch (Exception ignored1) {
                //ignored
                System.out.println("WARNING: rs.close() Exception: " + ignored1);
                ignored1.printStackTrace();
            }

            try {
                if(stmt != null)
                    stmt.close();
            } catch (Exception ignored2) {
                //ignored
                System.out.println("WARNING: stmt.close() Exception: " + ignored2);
                ignored2.printStackTrace();
            }
            //do not close connection: re-use for the next SQL in the transaction
        }

        return objectsFromDb;
    }


    ////////////////////// STANDARD GETTERS ///////////////////////

    /**
     * Method needed to implement Cacheable interface.
     */
    public String getId() {
        return this.id;
    }

    public String getTitle() {
        return this.title;
    }

    public String getArtist() {
        return this.artist;
    }

    public String getImageUrl() {
        return this.imageUrl;
    }

    public int getRemainingInInventory() {
        return this.remainingInInventory;
    }

    public double getPrice() {
        return this.price;
    }


    /**
     * Method needed to implement Cacheable interface.
     */
    public long getTimestamp() {
        return this.timestamp;
    }



    /////////////////////////////  UTILS  ////////////////////////////

    /**
     * Override the standard toString() 
     * with something remotely intelligent
     */
    public String toString() {
        return new StringBuffer()
            .append("CD {")
            .append(id)
            .append(", '")
            .append(title)
            .append("' by ")
            .append(artist)
            .append(". Image: ")
            .append(imageUrl)
            .append("; Itmes remaining: ")
            .append(remainingInInventory)
            .append(" @ $")
            .append(price)
            .append("; timestamp = ")
            .append(timestamp)
            .append("}")
            .toString();
    }
}



