---------------------------------------------------------------------------
License 
Copyright 2004 Grigory (Greg) Nudelman. All Rights Reserved. 
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: 

Redistribution of source code must retain this copyright notice, this list of conditions and the following disclaimer.

Redistribution in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. 

The name of Grigory (Greg) Nudelman may not be used to endorse or promote products derived from this software without specific prior written permission. 

This software is provided "AS IS," without a warranty of any kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. THE INVENTOR, Grigory (Greg) Nudelman AND THE LICENSORS OF THIS SOFTWARE SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL THE INVENTOR OR HIS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF THE INVENTOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. 

You acknowledge that this software is not designed, licensed or intended for use in the design, construction, operation or maintenance of any nuclear facility.  
---------------------------------------------------------------------------

package cache;

import java.sql.*;
import java.util.*;
import java.lang.reflect.*; 


/**
 * A static class that sets up, configures and manages the <code>Cache</code>
 * and optimized dynamic loading of <code>Timestamped</code> and 
 * <code>Cacheable</code> objects.<p>
 * 
 * The user can configure the cache size parameter and updater thread 
 * priority in two ways: 
 * <p>1) Pass in the property to the JVM: <p> 
 * 
 * > java -Dcachesize=15 -Dcacheupdaterpriority=10 cache.CD
 * 
 * <p>2) Create a cache.properties file in the classpath that contains lines: <p>
 * 
 * cachesize=15
 * cacheupdaterpriority=10
 *
 * <p>If this property is not found or can't be parsed correctly, 
 * we'll use the default cache size of 10, and default priority of 1 (MIN).
 * @author Greg Nudelman <GN>
 * @version 1.0
 */
public final class CacheManager {

    //constants
    private static final int    DEFAULT_CACHE_SIZE                = 10;
    private static final String CACHE_CONFIG_FILE_NAME            = "cache.properties";
    private static final String CACHE_SIZE_CONFIG_KEY             = "cachesize";
    private static final String CACHE_UPDATER_THREAD_PRIORITY_KEY = "cacheupdaterpriority";

    //our main cache we're managing here   
    private static Cache cache;

    //priority of the cache updater thread (1 low to 10 high)
    //no need to synchronize this one:
    //after intial set-up it is a read-only parameter
    private static int cacheUpdaterPriority; 

    //A private constructor will prevent anyone from manufacturing an instance.
    private CacheManager() {
    }

    /**
     * Try to find the cachesize property on command line, or  
     * in a cache.properties in the CLASSPATH
     * If cachesize is provided, try to parse it into an int, then set 
     * the cache to be that size. If cachesize property is not provided, 
     * (or it can't be parsed) leave it as a default value, 10.
     * Do the same thing with the cacheUpdaterPriority parameter. 
     * The value should be between 1 and 10 (inclusive). Defalt is 1.
     */
    static {
        synchronized(CacheManager.class) { //static synchronization
            try {
                //try to load the cache.properties file in 
                //the classpath into System properties object. 
                //Note: this will override the default System properties.
                System.getProperties().load(Thread.currentThread().
                                            getContextClassLoader().
                                            getResourceAsStream(CACHE_CONFIG_FILE_NAME));
            } catch(Exception ignore1) {
                //no big deal: we'll still try to get the cachesize param from the command-line options
                System.out.println("CacheManager properties file " + CACHE_CONFIG_FILE_NAME + 
                                   " is invalid or does not exist. Ignoring...");
            }


            try {
                //get the cachesize property
                int cacheSize = Integer.parseInt( ((String)System.getProperty(CACHE_SIZE_CONFIG_KEY)).trim() );
                System.out.println("CacheManager recovered a valid cacheSize from System properties: " + cacheSize);
                cache = new LRULinkedTableCache(cacheSize); //note: you can make the actual cache class name configurable the same way
                
            } catch (Exception ignore2) {
                
                //Can't find property, or can't parse it...whatever. Keep the default cache size: 10
                System.out.println("WARNING: CacheManager can't get cacheSize from System properties. Exception: " + 
                                   ignore2 + " Ignoring and using a DEFAULT CACHE SIZE: " + DEFAULT_CACHE_SIZE);
                cache = new LRULinkedTableCache(DEFAULT_CACHE_SIZE);
            }


            try {
                //get the cacheUpdaterPriority property
                cacheUpdaterPriority = Integer.parseInt( ((String)System.getProperty(CACHE_UPDATER_THREAD_PRIORITY_KEY)).trim() );
                System.out.println("CacheManager recovered a cacheUpdaterPriority from System properties: " + 
                                   cacheUpdaterPriority);
            } catch (Exception ignore3) {                
                //Can't find property, or can't parse it...whatever. Keep the default priority: 1
                System.out.println("CacheManager can't get cacheUpdaterPriority from System properties. " + 
                                   "Using default updater thread priority: " + Thread.MIN_PRIORITY);
                cacheUpdaterPriority = Thread.MIN_PRIORITY;
            }

            //validate what we got
            if(cacheUpdaterPriority > Thread.MAX_PRIORITY || cacheUpdaterPriority < Thread.MIN_PRIORITY) {
                throw new IllegalArgumentException("CacheManager's loaded cacheUpdaterPriority IS OUT OF RANGE: " + 
                                                   cacheUpdaterPriority + 
                                                   ". In the cache.properties file or java -D runtime params please set " +
                                                   CACHE_UPDATER_THREAD_PRIORITY_KEY +
                                                   " attribute to a value from 1 to 10 (inclusive). " + 
                                                   " (Or don't specify anything, to use the default value.)");
            }                                
        } //end of the static synchronized block
    }





    /**
     * Workhorse: Cache-optimized batch object loader 
     * of <code>Cacheable</code> objects. This variant Uses <code>Cacheable</code> 
     * Interface features to work on a passed instance of a <code>Cacheable</code> 
     * object to run the loading code. This is a more structured, preferred method 
     * to load <code>Cacheable</code> objects<p>
     * 
     * Will examine the cache for the passed <code>TimestampedId.id</code>, 
     * and if the cache does have the corresponding objects,
     * will compare the <code>timestamp</code> attribute of cached object with that 
     * provided by a <code>TimestampedId</code> object. If cache does not 
     * have it, or object is found to be stale, this method will 
     * (re-)load all the missing <code>Cacheable</code> objects from the DB 
     * as a group.<p>
     *
     * Detailed method sequence(using sequence numbers from the paper):<p> 
     * 3) Using the <code>Id:Timestamp</code> combination, go through our cache
     *    to collect "fresh" usable objects found in the cache 
     *    (for which the <code>Timestamp</code> has not changed).<p>
     * 4) At the same time, collect <code>Ids</code> for which either the <code>Timestamp</code>
     *    has been changed (object has been updated in the DB, 
     *    so the cached object is now "Stale"), or <code>Ids</code> that have not 
     *    been found in the cache at all (object is "Missing" 
     *    from the cache).<p>  
     * 5) Go back to our DB, load the data, and create any missing or 
     *    stale objects for all the <code>Ids</code> in step 4.<p> 
     * 6) Load all these updated and missing "new" and "re-loaded" 
     *    <code>Timestamped</code> objects from step 5 into our cache. 
     *    (These are "updates" to our cache.)<p> 
     * 7) Finally, combine a list from step 3 ("fresh" objects 
     *    retrieved from the cache) with the collection from step 5 
     *    ("new" and "re-loaded" objects from DB).<p>
     * 
     * Note that the final <code>List</code> comes back completely and randomly unordered,
     * and should be sorted in the caller method. Also an important point
     * is that you are not guaranteed successful load of <i>any</i> of the objects.
     * So, if you desire to ensure that the contents of the returned <code>List</code> 
     * match the number of the passed ids, you have to again check it
     * in the caller method.  This solution is more flexible, given potential 
     * dynamic nature of the <code>Cacheable</code> object that may be deleted and/or no 
     * longer available.
     *
     * @param timestampedIds
     *   <code>List</code> of <code>TimestampedId</code> objects for which complete fresh 
     *   objects need to be loaded from cache and/or DB.
     * @param cacheable
     *   an instance of a <code>Cacheable</code> object on which batch loading will be run.
     * @param conn
     *   a valid DB connection if one is needed for batch loading, or null
     * @return unordered <code>List</code> of fully-loaded, up-to-date (fresh) 
     *         <code>Timestamped</code> objects, from both <code>Cache</code> and/or DB, 
     *         that match passed timestampedIds list.
     * @exception ResourceLoadException in case of critical error
     * @author Greg Nudelman <GN>
     * @version 1.0
     */
    public static List loadFromCacheAndDB(List timestampedIds, Cacheable cacheable, Connection conn) throws ResourceLoadException {
        if(timestampedIds.isEmpty()) {
            System.out.println("loadFromCacheAndDB() nothing to load. Returning empty.");
            return new Vector(1);
        }

        Vector currentObjectsFromCache = new Vector(timestampedIds.size());
        Vector idsToRefreshFromDb      = new Vector(timestampedIds.size()); //String ids

        fillFreshAndStaleFromCache(timestampedIds, idsToRefreshFromDb, currentObjectsFromCache);

        Collection objectsFromDb = cacheable.loadFromIds(idsToRefreshFromDb, conn);

        updateCache(objectsFromDb);

        //combine the 2 collections: Timestamped's from cache and Timestamped's from DB
        currentObjectsFromCache.addAll(objectsFromDb);
        System.out.println("loadFromCacheAndDB() Got total from cache and DB: " + 
                              + currentObjectsFromCache.size());

        return currentObjectsFromCache; // return a vector of completely loaded, fresh Timestamped objects
    }


    /**
     * Variant: Cache-optimized batch object loader 
     * of <code>Timestamped</code> objects. This variant of the method
     * uses <i>Java Reflection API</i> to load the <code>Timestamped</code> 
     * objects from some unknown class's static method, all determined at runtime. 
     * This is a less preferred variant, since it represents a sort of "brute force" 
     * approach and uses only a sub-set of the <code>Timestamped</code> interface, 
     * the <code>Timestamped</code> interface. A better one to use is a more complete
     * "neater" <code>Cacheable</code> interface based aproach, 
     * but sometimes we find it necesary to load objects from existing classes 
     * that don't fit neatly into our new framework.  This method offers a convenient way out.<p>
     *
     * @see this.loadFromCacheAndDB(Cacheable)
     * @param timestampedIds
     *   <code>List</code> of <code>TimestampedId</code> objects for which complete fresh 
     *   objects need to be loaded from cache and/or DB.
     * @param loaderClassName
     *   a name of the class that has a static method to load the <code>Timestamped</code> objects from DB
     * @param loaderMethodName
     *   a name of the method that will load the objects from DB. <i>This Collection is completely un-ordered.</i>
     * @param conn
     *   a valid DB connection if one is needed for batch loading, or null
     * @return unordered <code>List</code> of fully-loaded, up-to-date (fresh) 
     *         <code>Timestamped</code> objects, from both <code>Cache</code> and/or DB, 
     *         that match passed timestampedIds list.
     * @exception ResourceLoadException in case of critical error
     * @author Greg Nudelman <GN>
     * @version 1.0
     */
    public static List loadFromCacheAndDB(List timestampedIds, String loaderClassName, String loaderMethodName, Connection conn) throws ResourceLoadException {
        if(timestampedIds.isEmpty()) {
            System.out.println("loadFromCacheAndDB() nothing to load. Returning empty.");
            return new Vector(1);
        }

        Vector currentObjectsFromCache = new Vector(timestampedIds.size());
        Vector idsToRefreshFromDb      = new Vector(timestampedIds.size()); //String ids

        fillFreshAndStaleFromCache(timestampedIds, idsToRefreshFromDb, currentObjectsFromCache);

        Collection objectsFromDb = loadStaleOrMissingFromDB(idsToRefreshFromDb, loaderClassName, loaderMethodName, conn);

        updateCache(objectsFromDb);

        //combine the 2 collections: Timestamped's from cache and Timestamped's from DB
        currentObjectsFromCache.addAll(objectsFromDb);
        System.out.println("loadFromCacheAndDB() Got total from cache and DB: " + 
                              + currentObjectsFromCache.size());

        return currentObjectsFromCache; // return a vector of completely loaded, fresh Timestamped objects
    }


    /**
     * A private helper method that will examine the cache for the presense 
     * of the id objects passed in the <code>timestampedIds List</code>.
     * Then, <p>
     * 3) Using the <code>Id:Timestamp</code> combination, go through our cache
     *    to collect "fresh" usable objects found in the cache 
     *    (for which the <code>Timestamp</code> has not changed).<p>
     * 4) At the same time, collect <code>Ids</code> for which either the <code>Timestamp</code>
     *    has been changed (object has been updated in the DB, 
     *    so the cached object is now "Stale"), or <code>Ids</code> that have not 
     *    been found in the cache at all (object is "Missing" 
     *    from the cache).<p>  
     * The two collections (fresh and stale) the method must fill are also passed in as parameters.
     * This is a somewhat funky design, since we can't return 2 objects and still want to have code re-use.
     * 
     * @author Greg Nudelman
     * @version 1.0
     * @param timestampedIds
     *   <code>List</code> of <code>TimestampedId</code> objects for which complete fresh 
     *   objects need to be loaded from cache and/or DB.
     * @param idsToRefreshFromDb
     *   empty Vector that will be filled by the stale and/or missing <i>ids</i> that should be refreshed from the DB
     * @param currentObjectsFromCache
     *   empty Vector that will be filled by the fresh up-to-date complete <i>objects</i>.
     */
    private static void fillFreshAndStaleFromCache(List timestampedIds, Vector idsToRefreshFromDb, Vector currentObjectsFromCache) {
        TimestampedId timestampedId = null;
        Timestamped c = null;
        for(Iterator i = timestampedIds.iterator(); i.hasNext(); ) {
            timestampedId = (TimestampedId)i.next();
            c = (Timestamped)cache.get(timestampedId.getId());

            if(null == c) {
                System.out.println("fillFreshAndStaleFromCache() '" + 
                                      timestampedId.getId() + "' <NOT FOUND> in cache");
                idsToRefreshFromDb.add(timestampedId.getId()); //String id
                continue;
            } else if(c.getTimestamp() != timestampedId.getTimestamp()) {
                System.out.println("fillFreshAndStaleFromCache() '" + 
                                      timestampedId.getId() + "' cached is <STALE> cache's " + c.getTimestamp() + 
                                      "!= id's " + timestampedId.getTimestamp());
                idsToRefreshFromDb.add(timestampedId.getId()); //String id
                continue;
            } else {
                System.out.println("fillFreshAndStaleFromCache() '" + 
                                      timestampedId.getId() + "' cached is <FRESH>");
                currentObjectsFromCache.add(c);
            }
        }
        
        System.out.println("fillFreshAndStaleFromCache() After looking in the cache:" + 
                              "\n currentObjectsFromCache = " + currentObjectsFromCache.size() + 
                              " ; idsToRefreshFromDb = " + idsToRefreshFromDb.size());

    }


    /**
     * A private helper method that will load the missing/stale objects not found 
     * in the cache from the DB. This variant of the method uses <i>Java Reflection API</i>
     * to load the object from some unknown class's static method, all determined at runtime. 
     * This is a less preferred variant, since it represents a sort of "brute force" 
     * approach and uses only a sub-set of the <code>Timestamped</code> interface, 
     * the <code>Timestamped</code> interface. A better one to use is a more complete
     * "neater" <code>Cacheable</code> interface based aproach, 
     * but sometimes we find it necesary to load objects from existing classes 
     * that don't fit neatly into our new framework.  This method offers a convenient way out.<p>
     *
     * Alternatively, this section is also constructed to call <code>loadFromIds()</code> on 
     * an instance of the <code>Cacheable</code> class (a <strong>NON-STATIC</strong> 
     * <code>loadFromIds</code> method is present in the <code>Cacheable</code> interface)
     * in this case our <code>loadFromIds</code> "loaderMethod" <strong>IS</strong> static, 
     * so since we can't define static methods in the interface, we'll have to invoke 
     * this method using reflection. <p>
     * 
     * 5) Go back to our DB, load the data, and create any missing or 
     *    stale objects for all the <code>Ids</code> in step 4.<p> 
     * 
     * @author Greg Nudelman
     * @version 1.0
     * @param idsToRefreshFromDb
     *   <code>List</code> of <code>String</code> ids for which complete fresh 
     *   objects need to be loaded from cache and/or DB.
     * @param loaderClassName
     *   a name of the class that has a static method to load the <code>Timestamped</code> objects from DB
     * @param loaderMethodName
     *   a name of the method that will load the objects from DB. <i>This Collection is completely un-ordered.</i>
     * @param conn
     *   a valid DB connection if one is needed for batch loading, or null
     * @return
     *   a <code>Collection</code> of new, fully-loaded objects
     * @exception ResourceLoadException
     *   in case of the invocation or SQL exceptions
     */
    private static Collection loadStaleOrMissingFromDB(Vector idsToRefreshFromDb, String loaderClassName, String loaderMethodName, Connection conn) 
        throws ResourceLoadException {

        Collection objectsFromDb = null;

        try {

            //(Re-)Load stale (or missing) Timestamped obj from DB
            Class loaderClass = Class.forName(loaderClassName);

            Class[] loaderMethodParameters = new Class[2];
            loaderMethodParameters[0] = Collection.class;  
            loaderMethodParameters[1] = Connection.class;

            Method loaderMethod = loaderClass.getMethod(loaderMethodName, loaderMethodParameters);
            Object[] loaderMethodArguments = new Object[2];
            loaderMethodArguments[0] = idsToRefreshFromDb;  
            loaderMethodArguments[1] = conn;

            objectsFromDb = (Collection)loaderMethod.invoke(null, loaderMethodArguments);  //null = static method

        } catch (Exception e) {
            if(e instanceof ResourceLoadException) { 
                throw (ResourceLoadException)e; //thrown by the loader method
            } else {
                System.out.println("Exception: loadStaleOrMissingFromDB() invalid method invocation: " + e);
                e.printStackTrace();
                throw new ResourceLoadException("Exception: loadStaleOrMissingFromDB() invalid class or method invocation: " + e);
            }
        }

        System.out.println("loadStaleOrMissingFromDB() Loaded from DB:" + 
                           "\n objectsFromDb = " + objectsFromDb.size());

        return objectsFromDb;
    }



    /**
     * update the cache with the incoming fresh <code>Timestamped</code> 
     * objects on a separate, lower-priority thread. Thread priority is 
     * a configurable parameter determined during the loading of this class.
     * Default value is Thread.MIN_PRIORITY.
     * 
     * @author Greg Nudelman
     * @version 1.0
     * @param freshTimestampedObjects
     *   a <code>Collection</code> of new, fully-loaded <code>Timestamped</code> objects
     */
    private static void updateCache(final Collection freshTimestampedObjects) {

        //put on a separate, low-priority thread
        Thread updater = new Thread() {
                public void run() {
                    System.out.println("START updateCache().run() updating with " + freshTimestampedObjects.size() + " CACHE CONTENTS: " + cache);
                    //update the cache with incoming fresh objects
                    Timestamped o = null;
                    for(Iterator i = freshTimestampedObjects.iterator(); i.hasNext(); ) {
                        o = (Timestamped)i.next();
                        cache.put(o.getId(), o);
                    }
                    System.out.println("END updateCache().run() CACHE CONTENTS: " + cache);
                }
            };

        updater.setPriority(CacheManager.cacheUpdaterPriority); //this is one parameter you may want to override in tunning the cache
        updater.start();
    }

}
