---------------------------------------------------------------------------
License 
Copyright 2004 Grigory (Greg) Nudelman. All Rights Reserved. 
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: 

Redistribution of source code must retain this copyright notice, this list of conditions and the following disclaimer.

Redistribution in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. 

The name of Grigory (Greg) Nudelman may not be used to endorse or promote products derived from this software without specific prior written permission. 

This software is provided "AS IS," without a warranty of any kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. THE INVENTOR, Grigory (Greg) Nudelman AND THE LICENSORS OF THIS SOFTWARE SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL THE INVENTOR OR HIS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF THE INVENTOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. 

You acknowledge that this software is not designed, licensed or intended for use in the design, construction, operation or maintenance of any nuclear facility.  
---------------------------------------------------------------------------

package cache;

import java.util.*;


/**
 * This class implements a fixed size Least-Recently-Used (LRU) cache.
 * The cache has a maximum size specified when it is created. When
 * an item is added to the cache, if the cache is already at the
 * maximum size the least recently used item is deleted, then the
 * new item is added.<p>
 * 
 * Thus the items cached are refered to by keys, just like a Hashtable,
 * and standard <code>java.util.Map</code> interface methods has been 
 * used in the API, to make using this class as similar to using
 * the standard Hashtable as possible.<p>
 *
 * The internal <code>java.util.LinkedHashMap</code> (JDK 1.4 and later) 
 * stores all the values, and does all the searching.
 * By accessing values by key through the underlying hashtable, 
 * we get access at constant-time.  Underlying linked list keeps track 
 * of the ordering and determines the "LRU value".<p>
 * 
 * Like <cache>HashMap</cache>, this class provides constant-time, or O(1), 
 * performance for the basic operations (<cache>add</cache>, 
 * <cache>contains</cache> and <cache>remove</cache>), 
 * assuming the the hash function disperses elements
 * properly among the buckets.  Performance is likely to be just slightly
 * below that of <cache>HashMap</cache>, due to the added expense of 
 * maintaining the linked list, with one exception: 
 * Iteration over the collection-views
 * of a <cache>LinkedHashMap</cache> requires time proportional to the <i>size</i>
 * of the map, regardless of its capacity.  Iteration over a <cache>HashMap</cache>
 * is likely to be more expensive, requiring time proportional to its
 * <i>capacity</i>.<p>
 * 
 * A linked hash map has two parameters that affect its performance:
 * <i>initial capacity</i> and <i>load factor</i>. 
 * The default load factor we chose for the cache 
 * (.75) offers a good tradeoff between time and space costs. 
 * Higher values decrease the space overhead but increase the lookup cost due 
 * to the higher number of collisions rather than direct lookups. 
 * If the initial capacity is greater than the maximum number of entries 
 * divided by the load factor, no time-consuming rehash operations will ever occur. 
 * Thus, to store a maximum of 100 elements in the cache with load of 0.75, 
 * we will size the internal map to 133.333 + 2 = (int)135.
 * (For a good discussion of this, see http://mindprod.com/jgloss/hashtable.html)<p>
 * 
 * <strong>Note that this implementation is not synchronized.</strong>
 * If this <code>Cache</code> will likely be used in a multi-threaded environment, 
 * a synchronized wrapper class <code>LRULinkedTableCache</code> 
 * should be used instead. Both cache classes are very similar in function, 
 * and both implement the <code>Cache</code> interface 
 * (which extends the <code>java.util.Map</code> interface.) 
 * This is similar to the standard 
 * <code>java.util.Hashtable</code> providing the synchronized wrapper for the 
 * <code>java.util.HashMap</code>.<p>
 * 
 * Implementation of this unsynchronized <code>LRULinkedMapCache</code>,
 * is best accomplished through the <code>Cache</code> interface, 
 * so that <code>LRULinkedTableCache</code> can be substituted 
 * with minimum of code modification if synchronization 
 * becomes an issue at a later date. Alternatively, you can use:<p> 
 * <code>
 * Map m = Collections.synchronizedMap(new LRULinkedMapCache(...));
 * </code>
 * <p>at creation time, to prevent accidental unsynchronized access.
 * 
 * @author Greg Nudelman
 * @version 1.0
 */ 
public class LRULinkedMapCache extends java.util.LinkedHashMap implements Cache {

    /*
     * Maximum allowed size of the cache.
     */
    private int maxSize;

    /*
     * The number of times get() failed to located the value in the cache
     * Mainly used for computing cache statistics
     */
    private long failedGetCounter;

    /*
     * The number of times get() successfully located the value in the cache 
     * Mainly used for computing cache statistics
     */
    private long successfulGetCounter;


    /**
     * Create a new, empty cache.
     * NOTE: in lower versions of JDK versions, prime numbers should make better 
     * hash sizes. However, in Java 1.4.1, for HashMap there no longer a payback
     * for specifying prime table lengths 
     * http://mindprod.com/jgloss/hashtable.html
     *
     * @param maxSize
     * The maxmimum number of values allowed in the cache.
     * When this limit is reached, the LRU value is deleted to
     * make room for the new value, so the total size of the cache
     * (the count of all the values it contains)
     * does not exceed the specified limit.
     */
    public LRULinkedMapCache(int maxSize) {
        super((maxSize * 4/3 + 2), 0.75f, true);
        if(maxSize <= 1)
            throw new IllegalArgumentException("LRULinkedMapCache(" + maxSize + 
                                               ") invalid cache size, must be > 1");

        this.maxSize = maxSize;
        clear();
    }



    //////////// Methods from interface java.util.Map ///////////////

    /**
     * Delete all the values from the cache.
     * Reset all the counters to 1. 
     * (Reseting to 1 will prevent /0 errors.
     * It will skew the statistics somewhat in the beginning,
     * though in the long run this should not matter much.)
     */
    public void clear() {
        super.clear();
        failedGetCounter = 1;     //this will prevent /0
        successfulGetCounter = 1; //this will prevent /0
    }


    /**
     * Fetch an value from the cache. Returns null if not found.
     * The fetched value becomes the "newest" value, and it's moved to the 
     * top of the linked list.
     * 
     * @param key
     *   The key used to refer to the object when <CODE>put()</CODE>
     *   was called.
     * @return
     *   The value, or null if not found.
     */
    public Object get(Object key) {
        Object value = super.get(key);
        if (value == null) {
            this.failedGetCounter++;
            return null; //not found
        } 

        //found
        this.successfulGetCounter++;
        return value;
    } 



    //////////// Methods from interface java.util.Map ///////////////

    /**
     * Returns the maxmimum number of values allowed in the cache.
     *
     * @return
     *   The maximum number of values allowed in the cache.
     */
    public int getMaxSize() {
        return maxSize;
    }



    ///////// Overriden methods from java.util.LinkedHashMap ///////////

    /**
     * Override this standard method for our cache size.
     * Returns true if this map should remove its eldest entry. 
     * This method is invoked by put and putAll after 
     * inserting a new entry into the map. 
     * 
     * @param eldest 
     *   The least recently inserted entry in the map, or if this is an 
     *   access-ordered map, the least recently accessed entry. 
     *   This is the entry that will be removed it this method returns true. 
     * @return
     *   true if the eldest entry should be removed from the map; 
     *   false if it should be retained.
     */
    protected boolean removeEldestEntry(Map.Entry eldest) {
        boolean remove = (size() > maxSize);
        return remove;
    }


  
    /////////////////////// Misc overriden methods /////////////////////

    /**
     * Warning: performance of this method is not that great 
     * (we have to iterate through all the values in the cache), 
     * so don't call this too often in production!
     *
     * @return a string describing the ordered contents of the cache,
     *   from least recently used, to most recently used.
     */
    public String toString() {

        StringBuffer buf = new StringBuffer()
            .append("\nLRULinkedMapCache ")
            .append(size())
            .append("/")
            .append(maxSize)
            .append(" | Failed: ")
            .append(failedGetCounter)
            .append(" | Succeesfull: ")
            .append(successfulGetCounter)
            .append(" | Missed: ")
            //assume under normal use: 
            //failedGet/(successGet + failedGet)
            .append((int)(   (  (double)failedGetCounter / 
                                ( (double)(successfulGetCounter) + 
                                  (double)failedGetCounter )  ) * 100   )) 
            .append("% Cache: { ");

        Collection keys = super.keySet();
        Iterator i = keys.iterator(); //sorted by LRU
        Object key = null;
        boolean isFirst = true;
        while (i.hasNext()) {
            if(! isFirst) {
                buf.append(", ");
            } else {
                isFirst = false;
            }

            key = i.next();

            buf.append(key);
            //warning: do not attempt to get the values straight from the hash
            //while running this iterator: 
            //concurrent modification exception will result
        }

        buf.append(" }\n");

        return buf.toString();
    }


    /**
     * Cloning not supported
     */
    public Object clone() {
        throw new UnsupportedOperationException();
    }



    /////////////////////// main() for testing  /////////////////////

    /**
     * Main is only for testing
     */
    public static void main(String[] args) {

        //gives Illegal Argument Exception (as it should)
        //LRULinkedMapCache f1 = new LRULinkedMapCache(-10);

        LRULinkedMapCache f = new LRULinkedMapCache(3);
                
        f.put("A", "VALUE");
        f.put("B", "VALUE");
        f.put("C", "VALUE");
        System.out.println(">>>>>>>>>>> 1 DONE <<<<<<<<<< " + f);

        f.get("C");
        f.get("C");
        f.get("C");
        f.get("C");
        f.get("C");
        f.get("C");

        f.get("A");
        f.get("A");
        f.get("A");

        f.get("B");
        f.get("B");
        f.get("junk");
        System.out.println(">>>>>>>>>>> 2 DONE <<<<<<<<<< " + f);

        f.put("D", "VALUE");
        f.put("E", "VALUE");
        f.put("F", "VALUE");
        f.put("A", "VALUE");

        System.out.println(">>>>>>>>>>> 3 DONE <<<<<<<<<< " + f);

        f.get("C");
        f.get("A");
        f.get("B");

        System.out.println(">>>>>>>>>>> 4 DONE <<<<<<<<<< " + f);

        ////////////10 objects/////////////
        LRULinkedMapCache f10 = new LRULinkedMapCache(10);
        Timer t = new Timer();
                
        t.start("store 10");
        for(int j = 0; j < 10; j++) {
            f10.put(new Integer(j), "DUD");
        }
        t.stop("store 10");

        t.start("10000 gets of 10 existing objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 0; j < 10; j++) {
                f10.get(new Integer(j));
            }
        }
        t.stop("10000 gets of 10 existing objects");

        t.start("10000 gets of 10 NON-existing objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 1000000; j < 1000010; j++) {
                f10.get(new Integer(j));
            }
        }
        t.stop("10000 gets of 10 NON-existing objects");


        t.start("10000 puts of new objects (remove LRU, put new obj)");
        for(int j = 10; j < 10010; j++) {
            f10.put(new Integer(j), "DUD");
        }
        t.stop("10000 puts of new objects (remove LRU, put new obj)");


        ////////////100 objects/////////////
        LRULinkedMapCache f100 = new LRULinkedMapCache(100);

        t.start("store 100");
        for(int j = 0; j < 100; j++) {
            f100.put(new Integer(j), "DUD");
        }
        t.stop("store 100");


        t.start("10000 gets of first 10 objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 0; j < 10; j++) {
                f100.get(new Integer(j));
            }
        }
        t.stop("10000 gets of first 10 objects");


        t.start("10000 gets of 10 NON-existing objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 1000000; j < 1000010; j++) {
                f100.get(new Integer(j));
            }
        }
        t.stop("10000 gets of 10 NON-existing objects");


        t.start("10000 puts of new objects (remove LRU, put new obj)");
        for(int j = 100; j < 10100; j++) {
            f100.put(new Integer(j), "DUD");
        }
        t.stop("10000 puts of new objects (remove LRU, put new obj)");


        ////////////1000 objects/////////////
        LRULinkedMapCache f1000 = new LRULinkedMapCache(1000);

        t.start("store 1000");
        for(int j = 0; j < 1000; j++) {
            f1000.put(new Integer(j), "DUD");
        }
        t.stop("store 1000");


        t.start("10000 gets of first 10 objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 0; j < 10; j++) {
                f1000.get(new Integer(j));
            }
        }
        t.stop("10000 gets of first 10 objects");


        t.start("10000 gets of 10 NON-existing objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 1000000; j < 1000010; j++) {
                f1000.get(new Integer(j));
            }
        }
        t.stop("10000 gets of 10 NON-existing objects");


        t.start("10000 puts of new objects (remove LRU, put new obj)");
        for(int j = 1000; j < 11000; j++) {
            f1000.put(new Integer(j), "DUD");
        }
        t.stop("10000 puts of new objects (remove LRU, put new obj)");


    }

}
