---------------------------------------------------------------------------
License 
Copyright 2004 Grigory (Greg) Nudelman. All Rights Reserved. 
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: 

Redistribution of source code must retain this copyright notice, this list of conditions and the following disclaimer.

Redistribution in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. 

The name of Grigory (Greg) Nudelman may not be used to endorse or promote products derived from this software without specific prior written permission. 

This software is provided "AS IS," without a warranty of any kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. THE INVENTOR, Grigory (Greg) Nudelman AND THE LICENSORS OF THIS SOFTWARE SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL THE INVENTOR OR HIS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF THE INVENTOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. 

You acknowledge that this software is not designed, licensed or intended for use in the design, construction, operation or maintenance of any nuclear facility.  
---------------------------------------------------------------------------

package cache;

import java.util.*;

/**
 * This class implements a fixed size Least-Recently-Used (LRU) cache.
 * The cache has a maximum size specified when it is created. When
 * an item is added to the cache, if the cache is already at the
 * maximum size the least recently used item is deleted, then the
 * new item is added.<p>
 * 
 * Thus the items cached are refered to by keys, just like a Hashtable,
 * and standard <code>java.util.Map</code> interface methods has been 
 * used in the API, to make using this class as similar to using
 * the standard Hashtable as possible.<p>
 *
 * The internal <code>java.util.LinkedHashMap</code> (JDK 1.4 and later) 
 * stores all the values, and does all the searching.
 * By accessing values by key through the underlying hashtable, 
 * we get access at constant-time.  Underlying linked list keeps track 
 * of the ordering and determines the "LRU value".<p>
 * 
 * Like <cache>HashMap</cache>, this class provides constant-time, or O(1), 
 * performance for the basic operations (<cache>add</cache>, 
 * <cache>contains</cache> and <cache>remove</cache>), 
 * assuming the the hash function disperses elements
 * properly among the buckets.  Performance is likely to be just slightly
 * below that of <cache>HashMap</cache>, due to the added expense of 
 * maintaining the linked list, with one exception: 
 * Iteration over the collection-views
 * of a <cache>LinkedHashMap</cache> requires time proportional to the <i>size</i>
 * of the map, regardless of its capacity.  Iteration over a <cache>HashMap</cache>
 * is likely to be more expensive, requiring time proportional to its
 * <i>capacity</i>.<p>
 * 
 * A linked hash map has two parameters that affect its performance:
 * <i>initial capacity</i> and <i>load factor</i>. 
 * The default load factor we chose for the cache 
 * (.75) offers a good tradeoff between time and space costs. 
 * Higher values decrease the space overhead but increase the lookup cost due 
 * to the higher number of collisions rather than direct lookups. 
 * If the initial capacity is greater than the maximum number of entries 
 * divided by the load factor, no time-consuming rehash operations will ever occur. 
 * Thus, to store a maximum of 100 elements in the cache with load of 0.75, 
 * we will size the internal map to 133.333 + 2 = (int)135.
 * (For a good discussion of this, see http://mindprod.com/jgloss/hashtable.html)<p>
 * 
 * <strong>This implementation is fully synchronized.</strong>
 * Since a cache will likely be used in a multi-threaded environment, 
 * this class provides a convenient synchronized wrapper for the 
 * <code>LRULinkedMapCache</code>. Just as the standard 
 * <code>java.util.Hashtable</code> provides the synchronized wrapper for the 
 * <code>java.util.HashMap</code>.
 * Both cache classes are very similar in function, and both implement the 
 * <code>Cache</code> interface (which extends the <code>java.util.Map</code> 
 * interface.)<p>
 * 
 * Implementation of the unsynchronized <code>LRULinkedMapCache</code>,
 * is best accomplished through the <code>Cache</code> interface, 
 * so that <code>LRULinkedTableCache</code> can be substituted 
 * with minimum of code modification if synchronization 
 * becomes an issue at a later date. Alternatively, you can use:<p> 
 * <code>
 * Map m = Collections.synchronizedMap(new LRULinkedMapCache(...));
 * </code>
 * <p>at creation time, to prevent accidental unsynchronized access.
 * 
 * @author Greg Nudelman
 * @version 1.0
 */ 
public class LRULinkedTableCache implements Cache {

    /*
     * The unsynchronized LRUCache that is wrapped by
     * the instance of this class.
     */
    private LRULinkedMapCache cache;

    /*
     * Maximum allowed size of the cache.
     */
    private int maxSize;


    /**
     * Create a new, empty cache.
     * NOTE: in lower versions of JDK versions, prime numbers should make 
     * better hash sizes. However, in Java 1.4.1, for HashMap there no 
     * longer a payback for specifying prime table lengths 
     * http://mindprod.com/jgloss/hashtable.html
     *
     * @param maxSize
     * The maxmimum number of values allowed in the cache.
     * When this limit is reached, the LRU value is deleted to
     * make room for the new value, so the total size of the cache
     * (the count of all the values it contains)
     * does not exceed the specified limit.
     */
    public LRULinkedTableCache(int maxSize) {
        if(maxSize <= 1)
            throw new IllegalArgumentException("LRULinkedTableCache(" + maxSize + 
                                               ") invalid cache size, must be > 1");

        cache = new LRULinkedMapCache(maxSize);

        this.maxSize = maxSize;
    }


    ////// additional Cache interface methods (not in Map interface) //////

    /**
     * Returns the maxmimum number of values allowed in the cache.
     *
     * @return
     *   The maximum number of values allowed in the cache.
     */
    public synchronized int getMaxSize() {
        return maxSize;
    }


    //////////// Methods from interface java.util.Map ///////////////

    /**
     * Delete all the mappings (values and keys) from this cache.
     */
    public synchronized void clear() {
        cache.clear();
    }


    /**
     * Does the cache contain the given key?
     *
     * @param key
     *   The key used to refer to the object when <code>add()</code>
     *   was called.
     * @return
     *   true if the key was found.
     */
    public synchronized boolean containsKey(Object key) {
        return cache.containsKey(key);
    }


    /**
     * Returns true if this map maps one or more keys to the specified value. 
     * @param value 
     *   value whose presence in this map is to be tested. 
     * @return
     *   true if this map maps one or more keys to the specified value.
     */
    public synchronized boolean containsValue(Object value) {
        return cache.containsValue(value);
    }


    /**
     * Returns a set view of the mappings contained in this map. 
     * Each element in the returned set is a Map.Entry. 
     * The set is backed by the map, so changes to the map are reflected 
     * in the set, and vice-versa. If the map is modified while an 
     * iteration over the set is in progress, the iterator will throw a 
     * <code>ConcurrentModificationException</code>. Thus, in the face 
     * of concurrent modification, the Iterator fails quickly and cleanly, 
     * rather than risking arbitrary, non-deterministic behavior 
     * at an undetermined time in the future. <p>
     * 
     * The set supports element removal, which removes the corresponding 
     * mapping from the map, via the <code>Iterator.remove</code>, 
     * <code>Set.remove</code>, <code>removeAll</code>, 
     * <code>retainAll</code> and <code>clear</code> operations. 
     * It does not support the <code>add</code> or <code>addAll</code> 
     * operations. 
     * 
     * @return
     *   a set view of the mappings contained in this map.
     */
    public synchronized Set entrySet() {
        return cache.entrySet();
    }


    /**
     * Fetch an value from the cache. Returns null if not found.
     * The fetched value becomes the "newest" value, and it's moved to the 
     * top of the linked list.
     * 
     * @param key
     *   The key used to refer to the object when <code>put()</code>
     *   was called.
     * @return
     *   The value, or null if not found.
     */
    public synchronized Object get(Object key) {
        return cache.get(key);
    } 


    /**
     * Returns true if this cache contains no key-value mappings. 
     *
     * @return
     *   true if this map contains no key-value mappings.
     */
    public synchronized boolean isEmpty() {
        return cache.isEmpty();
    }        


    /**
     * Returns a set view of the keys contained in this map. 
     * The set is backed by the map, so changes to the map are 
     * reflected in the set, and vice-versa. The set supports element 
     * removal, which removes the corresponding mapping from this map, 
     * via the Iterator.remove, Set.remove, removeAll, retainAll, 
     * and clear operations. It does not support the add or addAll operations. 
     *
     * @return
     *   a set view of the keys contained in this map.
     */
    public synchronized Set keySet() {
        return cache.keySet();
    }


    /**
     * Add an object to the cache.
     * This object becomes the Most-Recently-Used (MRU) item. 
     * If the cache is full, the LRU item will be deleted to make room. 
     * For the item that is already in the cache, 
     * put() will replace the value of the item and move it to the front of the
     * internal linked list (it will become the MRU item).
     * 
     * @param key
     *   The key used to refer to the object when calling <code>get()</code>.
     * @param item
     *   The item to add to the cache.
     */ 
    public synchronized Object put(Object key, Object value) {
        return cache.put(key, value);
    } 


    /**
     * Copies all of the mappings from the specified map to this cache. 
     * These mappings will replace any mappings that this cache had for
     * any of the keys currently in the specified map. 
     *
     * @param m 
     *   mappings to be stored in this map. 
     * @exception NullPointerException 
     *   if the specified map is null.
     */
    public synchronized void putAll(Map m) {
        cache.putAll(m);
    }


    /**
     * Removes the mapping for this key from this cache if it is present.
     *
     * @param key 
     *   key whose mapping is to be removed from the map. 
     * @return
     *   previous value associated with specified key, or null if there was no mapping for key. 
     */
    public synchronized Object remove(Object key) {
        return cache.remove(key);
    }


    /**
     * Returns the number of key-value mappings in this cache. 
     * 
     * @return 
     *   the number of key-value mappings in this map.
     */
    public synchronized int size() {
        return cache.size();
    }


    /**
     * Returns a collection view of the values contained in this map. 
     * The collection is backed by the map, so changes to the map are 
     * reflected in the collection, and vice-versa. 
     * If the map is modified while an iteration over the 
     * collection is in progress, the iterator will throw a 
     * <code>ConcurrentModificationException</code>. Thus, in the face 
     * of concurrent modification, the Iterator fails quickly and cleanly, 
     * rather than risking arbitrary, non-deterministic behavior 
     * at an undetermined time in the future. <p>
     * 
     * The collection supports element removal, which removes the corresponding 
     * mapping from the map, via the <code>Iterator.remove</code>, 
     * <code>Collection.remove</code>, <code>removeAll</code>, 
     * <code>retainAll</code> and <code>clear</code> operations. 
     * It does not support the <code>add</code> or <code>addAll</code> 
     * operations. 
     * 
     * @return
     *   a collection view of the values contained in this map.
     */
    public synchronized Collection values() {
        return cache.values();
    }


    /////////////////////// Misc overriden methods /////////////////////


    /**
     * Override the standard method with a more intelligent 
     * String represenation of the cache statistics and contents.
     * <p>
     * Warning: performance of this method is not that great 
     * (we have to iterate through all the values in the cache), 
     * so don't call this too often in production!
     *
     * @return a String describing stats and the ordered contents 
     *   of the cache, from least recently used, to most recently used.
     */
    public synchronized String toString() {

        return new StringBuffer()
            .append("\nLRULinkedTableCache (synchronized) { ")
            .append(cache.toString())
            .append("}")
            .toString();
    }


    /**
     * Cloning not supported
     */
    public Object clone() {
        throw new UnsupportedOperationException();
    }



    /////////////////////// main() for testing  /////////////////////

    /**
     * Main is only for testing
     */
    public static void main(String[] args) {

        //gives Illegal Argument Exception (as it should)
        //LRULinkedTableCache f1 = new LRULinkedTableCache(-10);

        LRULinkedTableCache f = new LRULinkedTableCache(3);
                
        f.put("A", "VALUE");
        f.put("B", "VALUE");
        f.put("C", "VALUE");
        System.out.println(">>>>>>>>>>> 1 DONE <<<<<<<<<< " + f);

        f.get("C");
        f.get("C");
        f.get("C");
        f.get("C");
        f.get("C");
        f.get("C");

        f.get("A");
        f.get("A");
        f.get("A");

        f.get("B");
        f.get("B");
        f.get("junk");
        System.out.println(">>>>>>>>>>> 2 DONE <<<<<<<<<< " + f);

        f.put("D", "VALUE");
        f.put("E", "VALUE");
        f.put("F", "VALUE");
        f.put("A", "VALUE");

        System.out.println(">>>>>>>>>>> 3 DONE <<<<<<<<<< " + f);

        f.get("C");
        f.get("A");
        f.get("B");

        System.out.println(">>>>>>>>>>> 4 DONE <<<<<<<<<< " + f);

        ////////////10 objects/////////////
        LRULinkedTableCache f10 = new LRULinkedTableCache(10);
        Timer t = new Timer();
                
        t.start("store 10");
        for(int j = 0; j < 10; j++) {
            f10.put(new Integer(j), "DUD");
        }
        t.stop("store 10");

        t.start("10000 gets of 10 existing objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 0; j < 10; j++) {
                f10.get(new Integer(j));
            }
        }
        t.stop("10000 gets of 10 existing objects");

        t.start("10000 gets of 10 NON-existing objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 1000000; j < 1000010; j++) {
                f10.get(new Integer(j));
            }
        }
        t.stop("10000 gets of 10 NON-existing objects");


        t.start("10000 puts of new objects (remove LRU, put new obj)");
        for(int j = 10; j < 10010; j++) {
            f10.put(new Integer(j), "DUD");
        }
        t.stop("10000 puts of new objects (remove LRU, put new obj)");


        ////////////100 objects/////////////
        LRULinkedTableCache f100 = new LRULinkedTableCache(100);

        t.start("store 100");
        for(int j = 0; j < 100; j++) {
            f100.put(new Integer(j), "DUD");
        }
        t.stop("store 100");


        t.start("10000 gets of first 10 objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 0; j < 10; j++) {
                f100.get(new Integer(j));
            }
        }
        t.stop("10000 gets of first 10 objects");


        t.start("10000 gets of 10 NON-existing objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 1000000; j < 1000010; j++) {
                f100.get(new Integer(j));
            }
        }
        t.stop("10000 gets of 10 NON-existing objects");


        t.start("10000 puts of new objects (remove LRU, put new obj)");
        for(int j = 100; j < 10100; j++) {
            f100.put(new Integer(j), "DUD");
        }
        t.stop("10000 puts of new objects (remove LRU, put new obj)");


        ////////////1000 objects/////////////
        LRULinkedTableCache f1000 = new LRULinkedTableCache(1000);

        t.start("store 1000");
        for(int j = 0; j < 1000; j++) {
            f1000.put(new Integer(j), "DUD");
        }
        t.stop("store 1000");


        t.start("10000 gets of first 10 objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 0; j < 10; j++) {
                f1000.get(new Integer(j));
            }
        }
        t.stop("10000 gets of first 10 objects");


        t.start("10000 gets of 10 NON-existing objects");
        for(int i = 0; i < 10000; i++) {
            for(int j = 1000000; j < 1000010; j++) {
                f1000.get(new Integer(j));
            }
        }
        t.stop("10000 gets of 10 NON-existing objects");


        t.start("10000 puts of new objects (remove LRU, put new obj)");
        for(int j = 1000; j < 11000; j++) {
            f1000.put(new Integer(j), "DUD");
        }
        t.stop("10000 puts of new objects (remove LRU, put new obj)");


    }

}
