---------------------------------------------------------------------------
License 
Copyright 2004 Grigory (Greg) Nudelman. All Rights Reserved. 
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: 

Redistribution of source code must retain this copyright notice, this list of conditions and the following disclaimer.

Redistribution in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. 

The name of Grigory (Greg) Nudelman may not be used to endorse or promote products derived from this software without specific prior written permission. 

This software is provided "AS IS," without a warranty of any kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. THE INVENTOR, Grigory (Greg) Nudelman AND THE LICENSORS OF THIS SOFTWARE SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL THE INVENTOR OR HIS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF THE INVENTOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. 

You acknowledge that this software is not designed, licensed or intended for use in the design, construction, operation or maintenance of any nuclear facility.  
---------------------------------------------------------------------------

package cache;

import java.io.*;
import java.sql.*;
import java.util.*;

/**
 * This class is a special runner/tester class for the Cache demo.
 * Note: This entire demo 
 * assumes that the id used in the system is a String.
 * If it's not, please adjust the code accordingly.
 *
 * @author  Greg Nudelman
 * @version 1.0
 * @see     CacheManager, CD
 */
public final class Tester {

    //do not instantiate
    private Tester() {}


    /**
     * Main will run this whole cache demo:
     * 1) Obtain a list of Ids for CDs we need to load, 
     *    based on the "Artist" parameter from std in.
     * 2) Retrieve the Timestamps of the objects to 
     *    which these Ids correspond.  
     * 3) Using the Id:Timestamp combination, go through our cache
     *    to collect "fresh" usable objects found in the cache 
     *    (for which the Timestamp has not changed).
     * 4) At the same time, collect Ids for which either the Timestamp
     *    has been changed (object has been updated in the DB, 
     *    so the cached object is now "Stale"), or Ids that have not 
     *    been found in the cache at all (object is "Missing" from the cache).  
     * 5) Go back to our DB, load the data, and create any missing or 
     *    stale objects for all the Ids in step 4. 
     * 6) Load all these updated and missing "new" and "re-loaded" 
     *    objects from step 5 into our cache. 
     *    (These are "updates" to our cache.) 
     * 7) Finally, combine collection from step 3 ("fresh" objects 
     *    retrieved from the cache) with the collection from step 5 
     *    ("new" and "re-loaded" objects from DB). 
     *
     * @author Greg Nudelman <GN>
     * @version 1.0
     */
    public static void main(String[] args) {
        
        //load the DB driver once, and init the CacheManager
        try {            
            //Class.forName("oracle.jdbc.driver.OracleDriver"); //Oracle
			Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
            Class.forName("cache.CacheManager"); //optional step to init the cache
        } catch(ClassNotFoundException cne) {                                           
            System.out.println("ClassNotFoundException: " + cne);
            cne.printStackTrace();
            return;
        }


        //get the search param from the user
        String artist = getArtistNameToSearchFor();
        while (artist != null) {
            if(artist.equals("")) {
                System.out.println("Bye!"); //exit the program, if just "return" was pressed
                break;
            }

            System.out.println("Searching for '" + artist + "'...");

            //we will be re-using this connection for the length of the user's "search" transaction
            Connection conn = null;

            try {            

                //open new connection for each search
                System.out.println("Openning DB connection."); 
				conn = DriverManager.getConnection("jdbc:odbc:cache",
                                                   "", 
                                                   "");

				//Edit this for your own Oracle Database
                //conn = DriverManager.getConnection("jdbc:oracle:thin:@sierra.tahoe.com:1521:PROD",
                //                                   "username", 
                //                                   "password");

                //1,2) get the timestamped ids for this artist
                List timestampedIds = CD.getCdIdsThatMatchUserSearch(artist, conn);
                if(! timestampedIds.isEmpty()) { 
                    System.out.println("Found " + timestampedIds.size() + " CDs for '" + artist + "'");
                    //3,4,5,6,7) found some CDs: try to load these objects from the cache and db

                    /***********************************************
                    //Batch-load using only the "Timestamped" interface, 
                    //by utilizing run-time invocation using Reflection API.
                    //To use this feature, CD only needs to implement 
                    //a more limited Timestamped interface. 
                    //the method loadFromIdsUsingReflection in cache.CD has to be static.
                    List cds = CacheManager.loadFromCacheAndDB(timestampedIds, "cache.CD", "loadFromIdsUsingReflection", conn);               
                    ***********************************************/

                    //batch-load using the Cacheable methods
                    List cds = CacheManager.loadFromCacheAndDB(timestampedIds, new CD(), conn);               

                    //sort by title to show to the user
                    performFinalSort(cds);

                    //print the final results
                    System.out.println("\n>>>>>>>>>> RESULTS OF YOUR SEARCH: >>>>>>>>>>\n");
                    System.out.println(Util.toString(cds));
                    System.out.println("\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n");
                    
                } else { 
                    //no CDs found. Abort loading and try another
                    System.out.println("Sorry, no CDs found for '" + artist + "'. Try another?");
                }

            } catch(ResourceLoadException ce) {                                                     
                System.out.println("ResourceLoadException: " + ce);
                ce.printStackTrace();
                return;
            } catch(SQLException sqle) { //can't get the connection                                                    
                System.out.println("SQLException: " + sqle);               
                sqle.printStackTrace();
                return;
            } finally {                                                                         
                //We're done with search transaction: close the connection.                            
                System.out.println("Search complete! Closing DB connection."); 
                try { 
                    if(conn != null) {                                                       
                        conn.close(); 
                        conn = null;
                    }
                } catch (Exception ignored) {
                    //ignored
                    System.out.println("WARNING: conn.close() Exception: " + ignored);
                    ignored.printStackTrace();
                }
            }        

            artist = getArtistNameToSearchFor(); //get the next search request
        }
                                                                 
    }


    ///////////////////// UTILITY METHODS //////////////////////

    /**
     * prints the instructions for the user to enter the artist's name
     * and accept the line from std in. If input error occurs, returns "" 
     * (that will exit the program)
     * @return String artist name to search for
     * @exception none
     * @version 1.0
     */
    private static String getArtistNameToSearchFor() {
        System.out.println("Please enter the artist name to search for, and press return." + 
                           " \n(To exit, just press return.) \nExample: >>>>Britney");
        System.out.print(">>>>");
        try {
            return new BufferedReader(new InputStreamReader(System.in)).readLine();
        } catch(IOException ioe) {                                                     
            System.out.println("IOException: " + ioe);
            return ""; //will exit the program
        }
    }


    /**
     * Sort the incoming <code>CD List</code> by title.
     * This is what we will be delivering to the user's UI.
     * @param cds
     *   List of CDs to sort by title
     */
    private static void performFinalSort(List cds) {

        Collections.sort(cds, 
                         //Order CDs by title
                         new Comparator() {
                             public int compare(Object cd1, Object cd2) {
                                 //HINT: -1 means cd1 is HIGHER rank
                                 return ( ((CD)cd1).getTitle() ).compareTo( ((CD)cd2).getTitle() );
                             }
                         }
                         );
    }


}
